@@ -118,13 +118,19 @@
 #define NXT_MTHD_METHODS(m)	(SCM_VELTS (m)[1])
 #define NXT_MTHD_ARGS(m)	(SCM_VELTS (m)[2])
 
+#ifdef SCM_UNBOUND
+#define SCM_GOOPS_UNBOUND SCM_UNBOUND
+#else
 #define SCM_GOOPS_UNBOUND scm_goops_the_unbound_value
-#define SCM_GOOPS_UNBOUNDP(x) ((x) == scm_goops_the_unbound_value)
+#endif
+#define SCM_GOOPS_UNBOUNDP(x) ((x) == SCM_GOOPS_UNBOUND)
 
 static int goops_loaded_p = 0;
 static scm_rstate *goops_rstate;
 
+#ifndef SCM_UNBOUND
 static SCM scm_goops_the_unbound_value;
+#endif
 static SCM scm_goops_lookup_closure;
 
 /* Some classes are defined in libguile/objects.c. */
@@ -149,7 +155,6 @@ SCM scm_class_int, scm_class_float, scm_class_double;
 
 SCM_SYMBOL (scm_sym_define_public, "define-public");
 
-static void set_slot_value_if_unbound (SCM class, SCM obj, SCM slot_name, SCM form);
 static SCM scm_make_unbound (void);
 static SCM scm_unbound_p (SCM obj);
 
@@ -371,6 +376,9 @@ SCM_PROC (s_sys_initialize_object, "%initialize-object", 2, 0, 0, scm_sys_initia
 
 SCM_KEYWORD (k_init_keyword, "init-keyword");
 
+static SCM get_slot_value (SCM class, SCM obj, SCM slotdef);
+static SCM set_slot_value (SCM class, SCM obj, SCM slotdef, SCM value);
+
 SCM
 scm_sys_initialize_object (SCM obj, SCM initargs)
 {
@@ -384,52 +392,68 @@ scm_sys_initialize_object (SCM obj, SCM initargs)
   SCM_ASSERT ((n_initargs & 1) == 0,
 	      initargs, SCM_ARG2, s_sys_initialize_object);
   
-  get_n_set = SCM_SLOT(class, scm_si_getters_n_setters);
-  slots     = SCM_SLOT(class, scm_si_slots);
+  get_n_set = SCM_SLOT (class, scm_si_getters_n_setters);
+  slots     = SCM_SLOT (class, scm_si_slots);
   
   /* See for each slot how it must be initialized */
-  for ( ; SCM_NNULLP(slots); get_n_set=SCM_CDR(get_n_set), slots=SCM_CDR(slots)) {
-    SCM slot_name  = SCM_CAR(slots);
-    SCM slot_value = 0;
+  for (;
+       SCM_NNULLP (slots);
+       get_n_set = SCM_CDR (get_n_set), slots = SCM_CDR (slots))
+    {
+      SCM slot_name  = SCM_CAR (slots);
+      SCM slot_value = 0;
     
-    if (SCM_NIMP (SCM_CDR (slot_name)))
-      {
-	/* This slot admits (perhaps) to be initialized at creation time */
-	int n = scm_ilength (SCM_CDR (slot_name));
-	if (n & 1) /* odd or -1 */
-	  scm_misc_error (s_sys_initialize_object,
-			  "class contains bogus slot definition: %S",
-			  SCM_LIST1 (slot_name));
-	tmp 	= scm_i_get_keyword (k_init_keyword,
+      if (SCM_NIMP (SCM_CDR (slot_name)))
+	{
+	  /* This slot admits (perhaps) to be initialized at creation time */
+	  int n = scm_ilength (SCM_CDR (slot_name));
+	  if (n & 1) /* odd or -1 */
+	    scm_misc_error (s_sys_initialize_object,
+			    "class contains bogus slot definition: %S",
+			    SCM_LIST1 (slot_name));
+	  tmp 	= scm_i_get_keyword (k_init_keyword,
 				     SCM_CDR (slot_name),
 				     n,
 				     0,
 				     s_sys_initialize_object);
-	slot_name = SCM_CAR(slot_name);
-	if (tmp) {
-	  /* an initarg was provided for this slot */
-	  if (!(SCM_NIMP (tmp) && SCM_KEYWORDP (tmp)))
-	    scm_misc_error (s_sys_initialize_object,
-			    "initarg must be a keyword. It was %S",
-			    SCM_LIST1 (tmp));
-	  slot_value = scm_i_get_keyword (tmp,
-					  initargs,
-					  n_initargs,
-					  0,
-					  s_sys_initialize_object);
+	  slot_name = SCM_CAR (slot_name);
+	  if (tmp)
+	    {
+	      /* an initarg was provided for this slot */
+	      if (!(SCM_NIMP (tmp) && SCM_KEYWORDP (tmp)))
+		scm_misc_error (s_sys_initialize_object,
+				"initarg must be a keyword. It was %S",
+				SCM_LIST1 (tmp));
+	      slot_value = scm_i_get_keyword (tmp,
+					      initargs,
+					      n_initargs,
+					      0,
+					      s_sys_initialize_object);
+	    }
 	}
-      }
 
-    if (slot_value)
-      /* set slot to provided value */
-      scm_slot_set_x (obj, slot_name, slot_value);
-    else {
-      /* set slot to its :init-form if it exists */
-      tmp = SCM_CADAR (get_n_set);
-      if (tmp != SCM_BOOL_F)
-	set_slot_value_if_unbound (class, obj, slot_name, tmp);
+      if (slot_value)
+	/* set slot to provided value */
+	set_slot_value (class, obj, SCM_CAR (get_n_set), slot_value);
+      else
+	{
+	  /* set slot to its :init-form if it exists */
+	  tmp = SCM_CADAR (get_n_set);
+	  if (tmp != SCM_BOOL_F)
+	    {
+	      slot_value = get_slot_value (class, obj, SCM_CAR (get_n_set));
+	      if (SCM_GOOPS_UNBOUNDP (slot_value))
+		{
+		  SCM env = SCM_EXTEND_ENV (SCM_EOL, SCM_EOL, SCM_ENV (tmp));
+		  set_slot_value (class,
+				  obj,
+				  SCM_CAR (get_n_set),
+				  scm_eval_body (SCM_CDR (SCM_CODE (tmp)),
+						 env));
+		}
+	    }
+	}
     }
-  }
   
   return obj;
 }
@@ -637,16 +661,7 @@ build_class_class_slots ()
 			      k_class,
 			      scm_class_self),
 	 scm_cons (Intern ("print"),
-	 scm_cons (SCM_LIST3 (Intern ("procedure0"),
-			      k_class,
-			      scm_class_protected_opaque),
-	 scm_cons (SCM_LIST3 (Intern ("procedure1"),
-			      k_class,
-			      scm_class_protected_opaque),
-	 scm_cons (SCM_LIST3 (Intern ("procedure2"),
-			      k_class,
-			      scm_class_protected_opaque),
-	 scm_cons (SCM_LIST3 (Intern ("procedure3"),
+	 scm_cons (SCM_LIST3 (Intern ("procedure"),
 			      k_class,
 			      scm_class_protected_opaque),
 	 scm_cons (SCM_LIST3 (Intern ("setter"),
@@ -689,7 +704,7 @@ build_class_class_slots ()
 	 scm_cons (Intern ("keyword-access"),
 	 scm_cons (Intern ("nfields"),
 	 scm_cons (Intern ("environment"),
-	 SCM_EOL)))))))))))))))))))))))))))))));
+	 SCM_EOL))))))))))))))))))))))))))));
 }
 
 static void
@@ -823,7 +838,7 @@ scm_class_precedence_list (SCM obj)
 {
   SCM_ASSERT (SCM_NIMP (obj) && CLASSP (obj),
 	      obj, SCM_ARG1, s_class_direct_precedence_list);
-  return scm_slot_ref(obj, Intern ("cpl"));
+  return scm_slot_ref (obj, Intern ("cpl"));
 }
 
 SCM_PROC (s_class_slots, "class-slots", 1, 0, 0, scm_class_slots);
@@ -833,7 +848,7 @@ scm_class_slots (SCM obj)
 {
   SCM_ASSERT (SCM_NIMP (obj) && CLASSP (obj),
 	      obj, SCM_ARG1, s_class_slots);
-  return scm_slot_ref(obj, Intern ("slots"));
+  return scm_slot_ref (obj, Intern ("slots"));
 }
 
 SCM_PROC (s_class_environment, "class-environment", 1, 0, 0, scm_class_environment);
@@ -920,6 +935,27 @@ scm_unbound_p (SCM obj)
   return SCM_GOOPS_UNBOUNDP (obj) ? SCM_BOOL_T : SCM_BOOL_F;
 }
 
+SCM_PROC (s_assert_bound, "assert-bound", 2, 0, 0, scm_assert_bound);
+
+static SCM
+scm_assert_bound (SCM value, SCM obj)
+{
+  if (SCM_GOOPS_UNBOUNDP (value))
+    return CALL_GF1 ("slot-unbound", obj);
+  return value;
+}
+
+SCM_PROC (s_at_assert_bound_ref, "@assert-bound-ref", 2, 0, 0, scm_at_assert_bound_ref);
+
+static SCM
+scm_at_assert_bound_ref (SCM obj, SCM index)
+{
+  SCM value = SCM_SLOT (obj, SCM_INUM (index));
+  if (SCM_GOOPS_UNBOUNDP (value))
+    return CALL_GF1 ("slot-unbound", obj);
+  return value;
+}
+
 SCM_PROC (s_sys_fast_slot_ref, "%fast-slot-ref", 2, 0, 0, scm_sys_fast_slot_ref);
 
 SCM
@@ -933,8 +969,7 @@ scm_sys_fast_slot_ref (SCM obj, SCM index)
   i = SCM_INUM (index);
   SCM_ASSERT (i >= 0 && i < SCM_NUMBER_OF_SLOTS (obj),
 	      index, SCM_OUTOFRANGE, s_sys_fast_slot_ref);
-
-  return SCM_SLOT (obj, i);
+  return scm_at_assert_bound_ref (obj, index);
 }
 
 SCM_PROC (s_sys_fast_slot_set_x, "%fast-slot-set!", 3, 0, 0, scm_sys_fast_slot_set_x);
@@ -945,137 +980,112 @@ scm_sys_fast_slot_set_x (SCM obj, SCM index, SCM value)
   register long i;
 
   SCM_ASSERT (SCM_NIMP (obj) && SCM_INSTANCEP (obj),
-	      obj, SCM_ARG1, s_sys_fast_slot_ref);
-  SCM_ASSERT (SCM_INUMP (index), index, SCM_ARG2, s_sys_fast_slot_ref);
+	      obj, SCM_ARG1, s_sys_fast_slot_set_x);
+  SCM_ASSERT (SCM_INUMP (index), index, SCM_ARG2, s_sys_fast_slot_set_x);
   i = SCM_INUM (index);
   SCM_ASSERT (i >= 0 && i < SCM_NUMBER_OF_SLOTS (obj),
-	      index, SCM_OUTOFRANGE, s_sys_fast_slot_ref);
+	      index, SCM_OUTOFRANGE, s_sys_fast_slot_set_x);
 
-  SCM_SLOT(obj, i) = value;
+  SCM_SLOT (obj, i) = value;
   return SCM_UNSPECIFIED;
 }
 
 /** Utilities **/
 
+/* In the future, this function will return the effective slot
+ * definition associated with SLOT_NAME.  Now it just returns some of
+ * the information which will be stored in the effective slot
+ * definition.
+ */
+
 static SCM
-get_slot_value (SCM class, SCM obj, SCM slot_name)
+slot_definition_using_name (SCM class, SCM slot_name)
 {
-  register SCM l;
-
-  for (l = SCM_ACCESSORS_OF (obj); SCM_NNULLP (l); l = SCM_CDR (l))
-    {
-      if (SCM_CAAR (l) == slot_name)
-	{
-	  l = SCM_CDDAR (l);
-	  /* Two cases here:
-	   *	- l is an integer (the offset of this slot in the slots vector)
-	   *	- otherwise (car l) is the getter function to apply
-	   */
-	  if (SCM_INUMP (l))
-	    return SCM_SLOT (obj, SCM_INUM (l));
-	  else
-	    {
-	      /* We must evaluate (apply (car l) (list obj)) 
-	       * where (car l) is known to be a closure of arity 1  */
-	      register SCM code, next, env;
-
-	      code = SCM_CAR (l);
-	      if (!SCM_CLOSUREP (code))
-		return SCM_SUBRF (code) (obj);
-	      env  = SCM_EXTEND_ENV (SCM_CAR (SCM_CODE (code)),
-				     SCM_LIST1 (obj),
-				     SCM_ENV (code));
-	      /* Evaluate the closure body */
-	      code = SCM_CDR (SCM_CODE (code));
-	    again:
-	      next = code;
-	      while (SCM_NNULLP (next = SCM_CDR (next)))
-		{
-		  if (SCM_IMP (SCM_CAR (code)))
-		    {
-		      if (SCM_ISYMP (SCM_CAR (code)))
-			{
-			  code = scm_m_expand_body (code, env);
-			  goto again;
-			}
-		    }
-		  else
-		    SCM_XEVAL (SCM_CAR (code), env);
-		  code = next;
-		}
-	      return SCM_XEVALCAR (code, env);
-	    }
-	}
-    }
-  return CALL_GF3 ("slot-missing", class, obj, slot_name);
+  register SCM slots = SCM_SLOT (class, scm_si_getters_n_setters);
+  for (; SCM_NIMP (slots); slots = SCM_CDR (slots))
+    if (SCM_CAAR (slots) == slot_name)
+      return SCM_CAR (slots);
+  return SCM_BOOL_F;
 }
 
 static SCM
-set_slot_value (SCM class, SCM obj, SCM slot_name, SCM value)
+get_slot_value (SCM class, SCM obj, SCM slotdef)
 {
-  register SCM l;
-
-  for (l = SCM_ACCESSORS_OF (obj); SCM_NNULLP (l); l = SCM_CDR (l))
-    {
-      if (SCM_CAAR (l) == slot_name)
-	{
-	  l = SCM_CDDAR (l);
-	  /* Two cases here:
-	   *	- l is an integer (the offset of this slot in the slots vector)
-	   *	- otherwise (cadr l) is the setter function to apply
+  SCM access = SCM_CDDR (slotdef);
+  /* Two cases here:
+   *	- access is an integer (the offset of this slot in the slots vector)
+   *	- otherwise (car access) is the getter function to apply
 	   */
-	  if (SCM_INUMP (l))
-	    SCM_SLOT (obj, SCM_INUM (l)) = value;
-	  else
-	    {
-	      /* We must evaluate (apply (cadr l) (list obj value))
-	       * where (cadr l) is known to be a closure of arity 2  */
-	      register SCM code, env;
+  if (SCM_INUMP (access))
+    return SCM_SLOT (obj, SCM_INUM (access));
+  else
+    {
+      /* We must evaluate (apply (car access) (list obj)) 
+       * where (car access) is known to be a closure of arity 1  */
+      register SCM code, env;
 
-	      code = SCM_CADR (l);
-	      if (!SCM_CLOSUREP (code))
-		SCM_SUBRF (code) (obj, value);
-	      else
-		{
-		  env  = SCM_EXTEND_ENV (SCM_CAR (SCM_CODE (code)),
-					 SCM_LIST2 (obj, value),
-					 SCM_ENV (code));
-		  /* Evaluate the closure body */
-		  code = SCM_CDR (SCM_CODE (code));
-		  while (SCM_NNULLP (code))
-		    {
-		      if (SCM_IMP (SCM_CAR (code)))
-			{
-			  if (SCM_ISYMP (SCM_CAR (code)))
-			    {
-			      code = scm_m_expand_body (code, env);
-			      continue;
-			    }
-			}
-		      else
-			SCM_XEVAL (SCM_CAR (code), env);
-		      code = SCM_CDR (code);
-		    }
-		}
-	    }
-	  return SCM_UNSPECIFIED;
+      code = SCM_CAR (access);
+      if (!SCM_CLOSUREP (code))
+	return SCM_SUBRF (code) (obj);
+      env  = SCM_EXTEND_ENV (SCM_CAR (SCM_CODE (code)),
+			     SCM_LIST1 (obj),
+			     SCM_ENV (code));
+      /* Evaluate the closure body */
+      return scm_eval_body (SCM_CDR (SCM_CODE (code)), env);
+    }
+}
+
+static SCM
+get_slot_value_using_name (SCM class, SCM obj, SCM slot_name)
+{
+  SCM slotdef = slot_definition_using_name (class, slot_name);
+  if (SCM_NFALSEP (slotdef))
+    return get_slot_value (class, obj, slotdef);
+  else
+    return CALL_GF3 ("slot-missing", class, obj, slot_name);
+}
+
+static SCM
+set_slot_value (SCM class, SCM obj, SCM slotdef, SCM value)
+{
+  SCM access = SCM_CDDR (slotdef);
+  /* Two cases here:
+   *	- access is an integer (the offset of this slot in the slots vector)
+   *	- otherwise (cadr access) is the setter function to apply
+   */
+  if (SCM_INUMP (access))
+    SCM_SLOT (obj, SCM_INUM (access)) = value;
+  else
+    {
+      /* We must evaluate (apply (cadr l) (list obj value))
+       * where (cadr l) is known to be a closure of arity 2  */
+      register SCM code, env;
+
+      code = SCM_CADR (access);
+      if (!SCM_CLOSUREP (code))
+	SCM_SUBRF (code) (obj, value);
+      else
+	{
+	  env  = SCM_EXTEND_ENV (SCM_CAR (SCM_CODE (code)),
+				 SCM_LIST2 (obj, value),
+				 SCM_ENV (code));
+	  /* Evaluate the closure body */
+	  scm_eval_body (SCM_CDR (SCM_CODE (code)), env);
 	}
     }
-  return CALL_GF4 ("slot-missing", class, obj, slot_name, value);
+  return SCM_UNSPECIFIED;
 }
 
-static void
-set_slot_value_if_unbound (SCM class, SCM obj, SCM slot_name, SCM form)
+static SCM
+set_slot_value_using_name (SCM class, SCM obj, SCM slot_name, SCM value)
 {
-  SCM old_val = get_slot_value (class, obj, slot_name);
-  
-  if (SCM_GOOPS_UNBOUNDP (old_val))
-    set_slot_value (class, obj,
-		    slot_name,
-		    scm_apply (form, SCM_EOL, SCM_EOL));
+  SCM slotdef = slot_definition_using_name (class, slot_name);
+  if (SCM_NFALSEP (slotdef))
+    return set_slot_value (class, obj, slotdef, value);
+  else
+    return CALL_GF4 ("slot-missing", class, obj, slot_name, value);
 }
 
-
 static SCM
 test_slot_existence (SCM class, SCM obj, SCM slot_name)
 {
@@ -1088,6 +1098,28 @@ test_slot_existence (SCM class, SCM obj, SCM slot_name)
   return SCM_BOOL_F;
 }
 
+/* The current libguile logand doesn't handle bignums.
+ * This (primitive) version handles them up to 32 bits.
+ */
+
+SCM_PROC1 (s_sys_logand, "%logand", scm_tc7_asubr, scm_sys_logand);
+
+static SCM
+scm_sys_logand (SCM n1, SCM n2)
+{
+  if (SCM_UNBNDP (n2))
+    {
+      if (SCM_UNBNDP (n1))
+	return SCM_MAKINUM (-1);
+      return n1;
+    }
+  {
+    unsigned long u1 = scm_num2ulong (n1, (char *) 1, s_sys_logand);
+    unsigned long u2 = scm_num2ulong (n2, (char *) 2, s_sys_logand);
+    return scm_ulong2num (u1 & u2);
+  }
+}
+
 		/* ======================================== */
 
 SCM_PROC (s_slot_ref_using_class, "slot-ref-using-class", 3, 0, 0, scm_slot_ref_using_class);
@@ -1104,7 +1136,7 @@ scm_slot_ref_using_class (SCM class, SCM obj, SCM slot_name)
   SCM_ASSERT (SCM_NIMP (slot_name) && SCM_SYMBOLP (slot_name),
 	      obj, SCM_ARG3, s_slot_ref_using_class);
 
-  res = get_slot_value (class, obj, slot_name);
+  res = get_slot_value_using_name (class, obj, slot_name);
   if (SCM_GOOPS_UNBOUNDP (res))
     return CALL_GF3 ("slot-unbound", class, obj, slot_name);
   return res;
@@ -1121,7 +1153,7 @@ scm_slot_set_using_class_x (SCM class, SCM obj, SCM slot_name, SCM value)
 	      obj, SCM_ARG2, s_slot_set_using_class_x);
   SCM_ASSERT (SCM_NIMP (slot_name) && SCM_SYMBOLP (slot_name),
 	      obj, SCM_ARG3, s_slot_set_using_class_x);
-  return set_slot_value (class, obj, slot_name, value);
+  return set_slot_value_using_name (class, obj, slot_name, value);
 }
 
 SCM_PROC (s_slot_bound_using_class_p, "slot-bound-using-class?", 3, 0, 0, scm_slot_bound_using_class_p);
@@ -1136,7 +1168,7 @@ scm_slot_bound_using_class_p (SCM class, SCM obj, SCM slot_name)
   SCM_ASSERT (SCM_NIMP (slot_name) && SCM_SYMBOLP (slot_name),
 	      obj, SCM_ARG3, s_slot_bound_using_class_p);
 
-  return (SCM_GOOPS_UNBOUNDP (get_slot_value (class, obj, slot_name))
+  return (SCM_GOOPS_UNBOUNDP (get_slot_value_using_name (class, obj, slot_name))
 	  ? SCM_BOOL_F
 	  : SCM_BOOL_T);
 }
@@ -1169,7 +1201,7 @@ scm_slot_ref (SCM obj, SCM slot_name)
 	      obj, SCM_ARG1, s_slot_ref);
   TEST_CHANGE_CLASS (obj, class);
 
-  res = get_slot_value (class, obj, slot_name);
+  res = get_slot_value_using_name (class, obj, slot_name);
   if (SCM_GOOPS_UNBOUNDP (res))
     return CALL_GF3 ("slot-unbound", class, obj, slot_name);
   return res;
@@ -1188,7 +1220,7 @@ scm_slot_set_x (SCM obj, SCM slot_name, SCM value)
 	      obj, SCM_ARG1, s_slot_set_x);
   TEST_CHANGE_CLASS(obj, class);
 
-  return set_slot_value (class, obj, slot_name, value);
+  return set_slot_value_using_name (class, obj, slot_name, value);
 }
 
 SCM_PROC (s_slot_bound_p, "slot-bound?", 2, 0, 0, scm_slot_bound_p);
@@ -1202,7 +1234,9 @@ scm_slot_bound_p (SCM obj, SCM slot_name)
 	      obj, SCM_ARG1, s_slot_bound_p);
   TEST_CHANGE_CLASS(obj, class);
 
-  return (SCM_GOOPS_UNBOUNDP (get_slot_value (class, obj, slot_name))
+  return (SCM_GOOPS_UNBOUNDP (get_slot_value_using_name (class,
+							 obj,
+							 slot_name))
 	  ? SCM_BOOL_F
 	  : SCM_BOOL_T);
 }
@@ -1230,8 +1264,6 @@ scm_slots_exists_p (SCM obj, SCM slot_name)
  *
  ******************************************************************************/
 
-static SCM f_apply_0_arity_method;
-
 static void clear_method_cache (SCM);
 
 static SCM
@@ -1293,10 +1325,7 @@ scm_sys_allocate_instance (SCM class, SCM initargs)
       else
 	{
 	  m[scm_struct_i_setter] = SCM_BOOL_F;
-	  m[scm_struct_i_proc + 0] = SCM_BOOL_F;
-	  m[scm_struct_i_proc + 1] = SCM_BOOL_F;
-	  m[scm_struct_i_proc + 2] = SCM_BOOL_F;
-	  m[scm_struct_i_proc + 3] = SCM_BOOL_F;
+	  m[scm_struct_i_procedure] = SCM_BOOL_F;
 	  return wrap_init (class, m, n);
 	}
     }
@@ -1411,10 +1440,72 @@ scm_sys_invalidate_class (SCM class)
   return SCM_UNSPECIFIED;
 }
 
-static void
-change_object_class (obj, old_class, new_class)
+/* When instances change class, they finally get a new body, but
+ * before that, they go through purgatory in hell.  Odd as it may
+ * seem, this data structure saves us from eternal suffering in
+ * infinite recursions.
+ */
+
+static SCM **hell;
+static int n_hell = 1;		/* one place for the evil one himself */
+static int hell_size = 4;
+#ifdef USE_THREADS
+static scm_mutex_t hell_mutex;
+#endif
+
+static int
+burnin (SCM o)
 {
-  CALL_GF3 ("change-object-class", obj, old_class, new_class);
+  int i;
+  for (i = 1; i < n_hell; ++i)
+    if (SCM_INST (o) == hell[i])
+      return i;
+  return 0;
+}
+
+static void
+go_to_hell (void *o)
+{
+#ifdef USE_THREADS
+  scm_mutex_lock (&hell_mutex);
+#endif
+  if (n_hell == hell_size)
+    {
+      int new_size = 2 * hell_size;
+      hell = scm_must_realloc (hell, hell_size, new_size, "hell");
+      hell_size = new_size;
+    }
+  hell[n_hell++] = SCM_INST (o);
+#ifdef USE_THREADS
+  scm_mutex_unlock (&hell_mutex);
+#endif
+}
+
+static void
+go_to_heaven (void *o)
+{
+#ifdef USE_THREADS
+  scm_mutex_lock (&hell_mutex);
+#endif
+  hell[burnin ((SCM) o)] = hell[--n_hell];
+#ifdef USE_THREADS
+  scm_mutex_unlock (&hell_mutex);
+#endif
+}
+
+static SCM
+purgatory (void *args)
+{
+  return scm_apply (GETVAR (Intern ("change-class")), (SCM) args, SCM_EOL);
+}
+
+static void
+change_class (obj, old_class, new_class)
+{
+  if (!burnin (obj))
+    scm_internal_dynamic_wind (go_to_hell, purgatory, go_to_heaven,
+			       (void *) SCM_LIST2 (obj, new_class),
+			       (void *) obj);
 }
 
 /******************************************************************************
@@ -1433,59 +1524,27 @@ change_object_class (obj, old_class, new_class)
  *
  ******************************************************************************/
 
-SCM_SYMBOL (sym_gf, "gf");
+SCM_KEYWORD (k_name, "name");
 
-static const char s_apply_0_arity_method[] = "apply-0-arity-method";
-static SCM var_compute_applicable_methods;
+SCM_SYMBOL (sym_no_method, "no-method");
 
-static SCM
-scm_apply_0_arity_method (SCM gf)
-{
-  SCM applicable = scm_apply (SCM_CDR (var_compute_applicable_methods),
-			      SCM_LIST2 (gf, SCM_EOL),
-			      SCM_EOL);
-  if (SCM_FALSEP (applicable))
-    {
-      CALL_GF2 ("no-applicable-method", gf, SCM_EOL);
-      return SCM_UNSPECIFIED;
-    }
-  else
-    {
-      SCM cmethod = CALL_GF2 ("compile-method", applicable, SCM_EOL);
-      SCM code;
-      SCM_ASSERT (SCM_NIMP (cmethod) && SCM_CONSP (cmethod)
-		  && SCM_NIMP (SCM_CDR (cmethod))
-		  && SCM_CONSP (SCM_CDR (cmethod)),
-		  cmethod,
-		  "strange cmethod returned by compile-method",
-		  s_apply_0_arity_method);
-      code = scm_acons (sym_gf,
-			SCM_EOL,
-			SCM_CDR (SCM_CMETHOD_CODE (cmethod)));
-      /* There can only be one 0-arity method. */
-      SCM_ENTITY_PROC_0 (gf) = scm_closure (code, SCM_CMETHOD_ENV (cmethod));
-      return scm_apply (gf, SCM_EOL, SCM_EOL);
-    }
-}
+static SCM list_of_no_method;
 
-static SCM sym_atdispatch, list_of_bool_f;
+SCM_SYMBOL (scm_sym_args, "args");
 
 SCM
 scm_make_method_cache (SCM gf)
 {
-  return SCM_LIST5 (sym_atdispatch, scm_sym_args, SCM_MAKINUM (1),
+  return SCM_LIST5 (SCM_IM_DISPATCH, scm_sym_args, SCM_MAKINUM (1),
 		    scm_make_vector (SCM_MAKINUM (SCM_INITIAL_MCACHE_SIZE),
-				     list_of_bool_f),
+				     list_of_no_method),
 		    gf);
 }
 
 static void
 clear_method_cache (SCM gf)
 {
-  SCM_ENTITY_PROC_0 (gf) = f_apply_0_arity_method;
-  SCM_ENTITY_PROC_1 (gf) = scm_make_method_cache (gf);
-  SCM_ENTITY_PROC_2 (gf) = scm_make_method_cache (gf);
-  SCM_ENTITY_PROC_3 (gf) = scm_make_method_cache (gf);
+  SCM_ENTITY_PROCEDURE (gf) = scm_make_method_cache (gf);
   SCM_SLOT (gf, scm_si_used_by) = SCM_BOOL_F;
 }
 
@@ -1509,9 +1568,56 @@ scm_sys_invalidate_method_cache_x (SCM gf)
 	   methods = SCM_CDR (methods))
 	SCM_SLOT (SCM_CAR (methods), scm_si_code_table) = SCM_EOL;
     }
+  SCM_SET_MCACHE_N_SPECIALIZED (SCM_ENTITY_PROCEDURE (gf),
+				SCM_SLOT (gf, scm_si_n_specialized));
   return SCM_UNSPECIFIED;
 }
 
+SCM_PROC (s_generic_capability_p, "generic-capability?", 1, 0, 0, scm_generic_capability_p);
+
+SCM
+scm_generic_capability_p (SCM proc)
+{
+  SCM_ASSERT (SCM_NFALSEP (scm_procedure_p (proc)),
+	      proc, SCM_ARG1, s_generic_capability_p);
+  return (scm_subr_p (proc) && SCM_SUBR_GENERIC (proc)
+	  ? SCM_BOOL_T
+	  : SCM_BOOL_F);
+}
+
+SCM_PROC (s_enable_primitive_generic_x, "enable-primitive-generic!", 0, 0, 1, scm_enable_primitive_generic_x);
+
+SCM
+scm_enable_primitive_generic_x (SCM subrs)
+{
+  while (SCM_NIMP (subrs))
+    {
+      SCM subr = SCM_CAR (subrs);
+      SCM_ASSERT (scm_subr_p (subr) && SCM_SUBR_GENERIC (subr),
+		  subr, SCM_ARGn, s_enable_primitive_generic_x);
+      *SCM_SUBR_GENERIC (subr)
+	= scm_make (SCM_LIST3 (scm_class_generic,
+			       k_name,
+			       SCM_SNAME (subr)));
+      subrs = SCM_CDR (subrs);
+    }
+  return SCM_UNSPECIFIED;
+}
+
+SCM_PROC (s_primitive_generic_generic, "primitive-generic-generic", 1, 0, 0, scm_primitive_generic_generic);
+
+SCM
+scm_primitive_generic_generic (SCM subr)
+{
+  if (scm_subr_p (subr) && SCM_SUBR_GENERIC (subr))
+    {
+      SCM gf = *SCM_SUBR_GENERIC (subr);
+      if (gf)
+	return gf;
+    }
+  return scm_wta (subr, (char *) SCM_ARG1, s_primitive_generic_generic);
+}
+
 /******************************************************************************
  * 
  * Protocol for calling a generic fumction
@@ -1685,7 +1791,12 @@ scm_compute_applicable_methods (SCM gf, SCM args, int len, int find_method_p)
   /* Build a list of all applicable methods */
   for (l = SCM_SLOT (gf, scm_si_methods); SCM_NNULLP (l); l = SCM_CDR (l))
     {
-      for (i = 0, fl = SPEC_OF (SCM_CAR (l)); ; i++, fl = SCM_CDR (fl))
+      fl = SPEC_OF (SCM_CAR (l));
+      /* Only accept accessors which match exactly in first arg. */
+      if (SCM_ACCESSORP (SCM_CAR (l))
+	  && (SCM_IMP (fl) || types[0] != SCM_CAR (fl)))
+	continue;
+      for (i = 0; ; i++, fl = SCM_CDR (fl))
 	{
 	  if ((SCM_NIMP (fl) && SCM_INSTANCEP (fl))
 	      /* We have a dotted argument list */
@@ -1756,53 +1867,32 @@ scm_m_atslot_set_x (SCM xorig, SCM env)
   return scm_cons (SCM_IM_SLOT_SET_X, x);
 }
 
-SCM_SYNTAX (s_dispatch, "@dispatch", scm_makmmacro, scm_m_dispatch);
+SCM_SYNTAX (s_atdispatch, "@dispatch", scm_makmmacro, scm_m_atdispatch);
 
-SCM_SYMBOL (sym_atdispatch, s_dispatch);
+SCM_SYMBOL (sym_atdispatch, s_atdispatch);
 
 SCM
-scm_m_dispatch (SCM xorig, SCM env)
+scm_m_atdispatch (SCM xorig, SCM env)
 {
   SCM args, n, v, gf, x = SCM_CDR (xorig);
-  SCM_ASSYNT (scm_ilength (x) == 4, xorig, scm_s_expression, s_dispatch);
+  SCM_ASSYNT (scm_ilength (x) == 4, xorig, scm_s_expression, s_atdispatch);
   args = SCM_CAR (x);
   SCM_ASSYNT (SCM_NIMP (args) && (SCM_CONSP (args) || SCM_SYMBOLP (args)),
-	      args, SCM_ARG1, s_dispatch);
+	      args, SCM_ARG1, s_atdispatch);
   x = SCM_CDR (x);
   n = SCM_XEVALCAR (x, env);
-  SCM_ASSYNT (SCM_INUMP (n), n, SCM_ARG2, s_dispatch);
-  SCM_ASSYNT (SCM_INUM (n) >= 1, n, SCM_OUTOFRANGE, s_dispatch);
+  SCM_ASSYNT (SCM_INUMP (n), n, SCM_ARG2, s_atdispatch);
+  SCM_ASSYNT (SCM_INUM (n) >= 1, n, SCM_OUTOFRANGE, s_atdispatch);
   x = SCM_CDR (x);
   v = SCM_XEVALCAR (x, env);
-  SCM_ASSYNT (SCM_NIMP (v) && SCM_VECTORP (v), v, SCM_ARG3, s_dispatch);
+  SCM_ASSYNT (SCM_NIMP (v) && SCM_VECTORP (v), v, SCM_ARG3, s_atdispatch);
   x = SCM_CDR (x);
   gf = SCM_XEVALCAR (x, env);
   SCM_ASSYNT (SCM_NIMP (gf) && SCM_STRUCTP (gf) && SCM_PUREGENERICP (gf),
-	      gf, SCM_ARG4, s_dispatch);
+	      gf, SCM_ARG4, s_atdispatch);
   return SCM_LIST5 (SCM_IM_DISPATCH, args, n, v, gf);
 }
 
-#if 0
-SCM_SYNTAX (s_hash_dispatch, "@hash-dispatch", scm_makmmacro, scm_m_hash_dispatch);
-
-SCM
-scm_m_hash_dispatch (SCM xorig, SCM env)
-{
-  SCM n, hashset, mask, v, x = SCM_CDR (xorig);
-  SCM_ASSYNT (scm_ilength (x) == 4, xorig, scm_s_expression, s_hash_dispatch);
-  n = SCM_XEVALCAR (x, env);
-  SCM_ASSYNT (SCM_INUMP (n), n, SCM_ARG1, s_dispatch);
-  x = SCM_CDR (x);
-  hashset = SCM_XEVALCAR (x, env);
-  mask = SCM_XEVALCAR (SCM_CDR (x), env);
-  v = SCM_XEVALCAR (SCM_CDDR (x), env);
-  SCM_ASSYNT (SCM_INUMP (hashset), hashset, SCM_ARG2, s_hash_dispatch);
-  SCM_ASSYNT (SCM_INUMP (mask), mask, SCM_ARG3, s_hash_dispatch);
-  SCM_ASSYNT (SCM_NIMP (v) && SCM_VECTORP (v), v, SCM_ARG4, s_hash_dispatch);
-  return SCM_LIST5 (SCM_IM_HASH_DISPATCH, n, hashset, mask, v);
-}
-#endif
-
 #ifdef USE_THREADS
 static void
 lock_cache_mutex (void *m)
@@ -1825,21 +1915,28 @@ call_memoize_method (void *a)
   SCM args = (SCM) a;
   SCM gf = SCM_CAR (args);
   SCM x = SCM_CADR (args);
+  /* First check if another thread has inserted a method between
+   * the cache miss and locking the mutex.
+   */
+  SCM cmethod = scm_mcache_lookup_cmethod (x, SCM_CDDR (args));
+  if (SCM_NIMP (cmethod))
+    return cmethod;
+  /*fixme* Use scm_apply */
   return CALL_GF3 ("memoize-method!", gf, SCM_CDDR (args), x);
 }
 
-static void
+static SCM
 memoize_method (SCM x, SCM args)
 {
   SCM gf = SCM_CAR (scm_last_pair (x));
 #ifdef USE_THREADS
-  scm_internal_dynamic_wind (lock_cache_mutex,
-			     call_memoize_method,
-			     unlock_cache_mutex,
-			     (void *) scm_cons2 (gf, x, args),
-			     (void *) SCM_SLOT (gf, scm_si_cache_mutex));
+  return scm_internal_dynamic_wind (lock_cache_mutex,
+				    call_memoize_method,
+				    unlock_cache_mutex,
+				    (void *) scm_cons2 (gf, x, args),
+				    (void *) SCM_SLOT (gf, scm_si_cache_mutex));
 #else
-  call_memoize_method ((void *) scm_cons2 (gf, x, args));
+  return call_memoize_method ((void *) scm_cons2 (gf, x, args));
 #endif
 }
 
@@ -1853,7 +1950,6 @@ memoize_method (SCM x, SCM args)
  *
  ******************************************************************************/
 
-SCM_KEYWORD (k_name,		"name");
 SCM_KEYWORD (k_setter,		"setter");
 SCM_KEYWORD (k_specializers,	"specializers");
 SCM_KEYWORD (k_procedure,	"procedure");
@@ -2095,7 +2191,7 @@ create_standard_classes (void)
 				     k_class,
 				     scm_class_opaque)));
   make_stdcls (&scm_class_foreign_object,  "<foreign-object>",
-	       scm_class_foreign_class, scm_class_top,	   SCM_EOL);
+	       scm_class_foreign_class, scm_class_object,   SCM_EOL);
   SCM_SET_CLASS_FLAGS (scm_class_foreign_object, SCM_CLASSF_FOREIGN);
 
   /* scm_class_generic functions classes */
@@ -2171,6 +2267,8 @@ create_standard_classes (void)
 	       scm_class_procedure_class, scm_class_top,   SCM_EOL);
   make_stdcls (&scm_class_procedure_with_setter, "<procedure-with-setter>",
 	       scm_class_procedure_class, scm_class_procedure, SCM_EOL);
+  make_stdcls (&scm_class_primitive_generic, "<primitive-generic>",
+	       scm_class_procedure_class, scm_class_procedure, SCM_EOL);
   make_stdcls (&scm_class_port,		   "<port>",
 	       scm_class_class, scm_class_top,		   SCM_EOL);
   make_stdcls (&scm_class_input_port,	   "<input-port>",
@@ -2296,7 +2394,6 @@ create_struct_classes (void)
   scm_internal_hash_fold (make_struct_class, 0, SCM_BOOL_F, scm_struct_table);
 }
 
-
 /**********************************************************************
  *
  * C interface
@@ -2320,8 +2417,6 @@ scm_make_foreign_object (SCM class, SCM initargs)
   return scm_wrap_object (class, constructor (initargs));
 }
 
-SCM_SYMBOL (sym_write_object, "write-object");
-
 static size_t
 scm_free_foreign_object (SCM *class, SCM *data)
 {
@@ -2355,8 +2450,6 @@ scm_make_class (SCM meta, char *s_name, SCM supers, size_t size,
   
   SCM_SLOT (class, scm_si_layout) = SCM_CAR (scm_intern ("", 0));
   SCM_SLOT (class, scm_si_constructor) = (SCM) constructor;
-  SCM_SLOT (class, scm_si_print) = scm_eval2 (sym_write_object,
-					      scm_goops_lookup_closure);
 
   return class;
 }
@@ -2517,32 +2610,37 @@ sys_goops_loaded ()
 void
 scm_init_goops (void)
 {
-  SCM the_unbound_value = SCM_CAR (scm_intern0 ("the-unbound-value"));
   SCM goops_module = scm_make_module (scm_read_0str ("(oop goops)"));
   SCM old_module = scm_select_module (goops_module);
+#ifndef SCM_UNBOUND
+  SCM the_unbound_value = SCM_CAR (scm_intern0 ("the-unbound-value"));
+  scm_goops_the_unbound_value
+    = scm_permanent_object (scm_cons (the_unbound_value, SCM_EOL));
+#endif
+  
   scm_goops_lookup_closure = scm_module_lookup_closure (goops_module);
-  scm_apply_generic_env
-    = scm_permanent_object (SCM_LIST1 (scm_goops_lookup_closure));
 
   scm_components = scm_permanent_object (scm_make_weak_key_hash_table
 					 (SCM_MAKINUM (37)));
 
   goops_rstate = scm_c_make_rstate ("GOOPS", 5);
 
-  f_apply_0_arity_method
-    = scm_make_gsubr (s_apply_0_arity_method, 1, 0, 0,
-		      scm_apply_0_arity_method);
-
 #include "goops.x"
 
-  list_of_bool_f = scm_permanent_object (SCM_LIST1 (SCM_BOOL_F));
-  
-  scm_goops_the_unbound_value
-    = scm_permanent_object (scm_cons (the_unbound_value, SCM_EOL));
-  
+  list_of_no_method = scm_permanent_object (SCM_LIST1 (sym_no_method));
+
+  hell = scm_must_malloc (hell_size, "hell");
+#ifdef USE_THREADS
+#ifdef SCM_MUTEX_INIT_TWO_ARGS
+  scm_mutex_init (&hell_mutex, NULL);
+#else
+  scm_mutex_init (&hell_mutex);
+#endif
+#endif
+
   scm_make_extended_class = make_extended_class;
   scm_make_port_classes = make_port_classes;
-  scm_change_object_class = change_object_class;
+  scm_change_object_class = change_class;
   scm_memoize_method = memoize_method;
 
   create_basic_classes ();
@@ -2551,6 +2649,15 @@ scm_init_goops (void)
   create_struct_classes ();
   create_port_classes ();
 
+  {
+    SCM name = SCM_CAR (scm_intern0 ("no-applicable-method"));
+    scm_no_applicable_method
+      = scm_permanent_object (scm_make (SCM_LIST3 (scm_class_generic,
+						   k_name,
+						   name)));
+    DEFVAR (name, scm_no_applicable_method);
+  }
+  
   scm_select_module (old_module);
 }
 
