@@ -42,25 +42,24 @@
     define-accessor make-accessor ensure-accessor
     define-method make-method method add-method!
     object-eqv? object-equal?
-    write-object display-object
     class-slot-ref class-slot-set! slot-unbound slot-missing 
     slot-definition-name  slot-definition-options slot-definition-allocation
     slot-definition-getter slot-definition-setter slot-definition-accessor
-    slot-definition-init-value
+    slot-definition-init-value slot-definition-init-form
     slot-definition-init-thunk slot-definition-init-keyword 
     slot-init-function class-slot-definition
     method-source
-    compute-cpl compute-std-cpl compute-get-n-set
+    compute-cpl compute-std-cpl compute-get-n-set compute-slots
+    compute-getter-method compute-setter-method
     allocate-instance initialize make-instance make
     no-next-method  no-applicable-method no-method
-    change-class 
+    change-class update-instance-for-different-class
     shallow-clone deep-clone
     class-redefinition
     apply-generic apply-method apply-methods
     compute-applicable-methods %compute-applicable-methods
     method-more-specific? sort-applicable-methods
     class-subclasses class-methods
-    slot-value
     goops-error
 )
 
@@ -72,9 +71,12 @@
     slot-exists-using-class? slot-ref slot-set! slot-bound? class-of
     class-name class-direct-supers class-direct-subclasses
     class-direct-methods class-direct-slots class-precedence-list
-    class-slots class-environment generic-function-name
+    class-slots class-environment
+    generic-function-name
     generic-function-methods method-generic-function method-specializers
-    method-procedure slot-exists? make find-method get-keyword)
+    primitive-generic-generic enable-primitive-generic!
+    method-procedure slot-exists? make find-method get-keyword
+    %logand)
 
 
 ;;
@@ -378,7 +380,8 @@
   (let ((name (and (pair? name) (car name))))
     (make <generic-with-setter>
 	  #:name name
-	  #:setter (make <generic> #:name (make-setter-name name)))))
+	  #:setter (make <generic>
+		         #:name (and name (make-setter-name name))))))
 
 (define (ensure-accessor proc . name)
   (let ((name (and (pair? name) (car name))))
@@ -427,8 +430,9 @@
 		     (goops-error "bad method name: %S" name))
 		    ((defined? name env)
 		     `(begin
-			(if (not (is-a? ,name <generic-with-setter>))
-			    (define-accessor ,name))
+			;; *fixme* Temporary hack for the current module system
+			(if (not ,name)
+			    (define-generic ,name))
 			(add-method! (setter ,name) (method ,@(cddr exp)))))
 		    (else
 		     `(begin
@@ -438,7 +442,8 @@
 		   (goops-error "bad method name: %S" name))
 		  ((defined? name env)
 		   `(begin
-		      (if (not (is-a? ,name <generic>))
+		      ;; *fixme* Temporary hack for the current module system
+		      (if (not ,name)
 			  (define-generic ,name))
 		      (add-method! ,name (method ,@(cddr exp)))))
 		  (else
@@ -514,8 +519,12 @@
 (define (internal-add-method! gf m)
   (slot-set! m  'generic-function gf)
   (slot-set! gf 'methods (compute-new-list-of-methods gf m))
-  (slot-set! gf 'n-specialized (max (length* (slot-ref m 'specializers))
-				    (slot-ref gf 'n-specialized)))
+  (let ((specializers (slot-ref m 'specializers)))
+    (slot-set! gf 'n-specialized
+	       (max (if (list? specializers)
+			(length specializers)
+			(+ 1 (length* specializers)))
+		    (slot-ref gf 'n-specialized))))
   (%invalidate-method-cache! gf)
   (add-method-in-classes! m)
   *unspecified*)
@@ -527,6 +536,19 @@
 			#:specializers (list <generic> <method>)
 			#:procedure internal-add-method!))
 
+(define-method add-method! ((proc <procedure>) (m <method>))
+  (if (generic-capability? proc)
+      (begin
+	(enable-primitive-generic! proc)
+	(add-method! proc m))
+      (next-method)))
+
+(define-method add-method! ((pg <primitive-generic>) (m <method>))
+  (add-method! (primitive-generic-generic pg) m))
+
+(define-method add-method! (obj (m <method>))
+  (goops-error "%S is not a valid generic function" obj))
+
 ;;;
 ;;; {Access to meta objects}
 ;;;
@@ -537,7 +559,7 @@
 (define-method method-source ((m <method>))
   (let* ((spec (map* class-name (slot-ref m 'specializers)))
 	 (proc (procedure-source (slot-ref m 'procedure)))
-	 (args (cdadr proc))
+	 (args (cadr proc))
 	 (body (cddr proc)))
     (cons 'method
 	  (cons (map* list args spec)
@@ -563,8 +585,12 @@
   (get-keyword #:accessor (cdr s) #f))
 
 (define (slot-definition-init-value s)
+  ;; can be #f, so we can't use #f as non-value
   (get-keyword #:init-value (cdr s) (make-unbound)))
 
+(define (slot-definition-init-form s)
+  (get-keyword #:init-form (cdr s) (make-unbound)))
+
 (define (slot-definition-init-thunk s)
   (get-keyword #:init-thunk (cdr s) #f))
 
@@ -600,12 +626,14 @@
 (define (display-address o file)
   (display (number->string (object-address o) 16) file))
 
-(define-method write-object (o file)
+(define-method write (o file)
   (display "#<instance " file)
   (display-address o file)
   (display #\> file))
 
-(define-method write-object ((o <object>) file)
+(define write-object (primitive-generic-generic write))
+
+(define-method write ((o <object>) file)
   (let ((class (class-of o)))
     (if (slot-bound? class 'name)
 	(begin
@@ -616,7 +644,7 @@
 	  (display #\> file))
 	(next-method))))
 
-(define-method write-object ((o <foreign-object>) file)
+(define-method write ((o <foreign-object>) file)
   (let ((class (class-of o)))
     (if (slot-bound? class 'name)
 	(begin
@@ -627,7 +655,7 @@
 	  (display #\> file))
 	(next-method))))
 
-(define-method write-object ((class <class>) file)
+(define-method write ((class <class>) file)
   (let ((meta (class-of class)))
     (if (and (slot-bound? class 'name)
 	     (slot-bound? meta 'name))
@@ -641,7 +669,7 @@
 	  (display #\> file))
 	(next-method))))
 
-(define-method write-object ((gf <generic>) file)
+(define-method write ((gf <generic>) file)
   (let ((meta (class-of gf)))
     (if (and (slot-bound? meta 'name)
 	     (slot-bound? gf 'methods))
@@ -658,7 +686,7 @@
 	  (display ")>" file))
 	(next-method))))
 
-(define-method write-object ((o <method>) file)
+(define-method write ((o <method>) file)
   (let ((meta (class-of o)))
     (if (and (slot-bound? meta 'name)
 	     (slot-bound? o 'specializers))
@@ -677,31 +705,8 @@
 	  (display #\> file))
 	(next-method))))
 
-;; Teach all created classes how to print themselves
-(for-each (lambda (class)
-	    (slot-set! class 'print write-object))
-	  (list <top> <object> <class>
-		<procedure-class> <entity-class>
-		<operator-class> <operator-with-setter-class>
-		<entity> <entity-with-setter>
-		<method> <simple-method> <accessor-method>
-		<generic> <generic-with-setter>
-		<boolean> <char>
-		<list> <pair> <null>
-		<string> <symbol>
-		<vector>
-		<number> <complex> <real> <integer>
-		<keyword>
-		<unknown>
-		<procedure>
-		<foreign-object> <foreign-class> <foreign-slot>
-		<protected-slot> <opaque-slot> <read-only-slot> <self-slot>
-		<protected-opaque-slot> <protected-read-only-slot>
-		<scm-slot> <int-slot> <float-slot> <double-slot>
-		))
-
 ;; Display (do the same thing as write by default)
-(define-method display-object (o file) 
+(define-method display (o file) 
   (write-object o file))
 
 ;;;
@@ -732,6 +737,9 @@
 (define-method slot-unbound ((c <class>) s)
   (goops-error "Slot `%S' is unbound in class %S" s c))
 
+(define-method slot-unbound ((o <object>))
+  (goops-error "Unbound slot in object %S" o))
+
 (define-method slot-missing ((c <class>) (o <object>) s)
   (goops-error "No slot with name `%S' in object %S" s o))
   
@@ -748,7 +756,7 @@
   (goops-error "No next method when calling %S\nwith arguments %S" gf args))
 
 (define-method no-applicable-method ((gf <generic>) args)
-  (goops-error "No applicable method for %S\nin call %S"
+  (goops-error "No applicable method for %S in call %S"
 	       gf (cons (generic-function-name gf) args)))
 
 (define-method no-method ((gf <generic>) args)
@@ -899,40 +907,57 @@
 	(let ((name            (slot-definition-name     s))
 	      (getter-function (slot-definition-getter   s))
 	      (setter-function (slot-definition-setter   s))
-	      (accessor        (slot-definition-accessor s))
-	      (g-n-s (cddr g-n-s)))
+	      (accessor        (slot-definition-accessor s)))
 	  (if getter-function
 	      (add-method! getter-function
-			   (make <accessor-method>
-				 #:specializers (list class)
-				 #:procedure (if (pair? g-n-s)
-						 (car g-n-s)
-						 (standard-get g-n-s)))))
+			   (compute-getter-method class g-n-s)))
 	  (if setter-function
 	      (add-method! setter-function
-			   (make <accessor-method>
-				 #:specializers (list class <top>)
-				 #:procedure (if (pair? g-n-s)
-						 (cadr g-n-s)
-						 (standard-set g-n-s)))))
+			   (compute-setter-method class g-n-s)))
 	  (if accessor
 	      (begin
 		(add-method! accessor
-			     (make <accessor-method>
-				   #:specializers (list class)
-				   #:procedure (if (pair? g-n-s)
-						 (car g-n-s)
-						 (standard-get g-n-s))))
+			     (compute-getter-method class g-n-s))
 		(add-method! (setter accessor)
-			     (make <accessor-method>
-				   #:specializers (list class <top>)
-				   #:procedure (if (pair? g-n-s)
-						 (cadr g-n-s)
-						 (standard-set g-n-s))))))))
+			     (compute-setter-method class g-n-s))))))
       slots (slot-ref class 'getters-n-setters)))
 
+(define-method compute-getter-method ((class <class>) slotdef)
+  (let ((init-thunk (cadr slotdef))
+	(g-n-s (cddr slotdef)))
+    (make <accessor-method>
+          #:specializers (list class)
+	  #:procedure (cond ((pair? g-n-s)
+			     (if init-thunk
+				 (car g-n-s)
+				 (make-generic-bound-check-getter (car g-n-s))
+				 ))
+			    (init-thunk
+			     (standard-get g-n-s))
+			    (else
+			     (bound-check-get g-n-s))))))
+
+(define-method compute-setter-method ((class <class>) slotdef)
+  (let ((g-n-s (cddr slotdef)))
+    (make <accessor-method>
+          #:specializers (list class <top>)
+	  #:procedure (if (pair? g-n-s)
+			  (cadr g-n-s)
+			  (standard-set g-n-s)))))
+
+(define (make-generic-bound-check-getter proc)
+  (let ((source (and (closure? proc) (procedure-source proc))))
+    (if (and source (null? (cdddr source)))
+	(let ((obj (caadr source)))
+	  ;; smart closure compilation
+	  (local-eval
+	   `(lambda (,obj) (assert-bound ,(caddr source) ,obj))
+	   (procedure-environment proc)))
+	(lambda (o) (assert-bound (proc o) o)))))
+
 (define n-standard-accessor-methods 10)
 
+(define bound-check-get-methods (make-vector n-standard-accessor-methods #f))
 (define standard-get-methods (make-vector n-standard-accessor-methods #f))
 (define standard-set-methods (make-vector n-standard-accessor-methods #f))
 
@@ -944,12 +969,17 @@
 		  (vector-set! methods index m)
 		  m)))))
 
+(define (make-bound-check-get index)
+  (local-eval `(lambda (o) (@assert-bound-ref o ,index)) (the-environment)))
+
 (define (make-get index)
   (local-eval `(lambda (o) (@slot-ref o ,index)) (the-environment)))
 
 (define (make-set index)
   (local-eval `(lambda (o v) (@slot-set! o ,index v)) (the-environment)))
 
+(define bound-check-get
+  (standard-accessor-method make-bound-check-get bound-check-get-methods))
 (define standard-get (standard-accessor-method make-get standard-get-methods))
 (define standard-set (standard-accessor-method make-set standard-set-methods))
 
@@ -1129,9 +1159,7 @@
      (let ((name (slot-definition-name s)))
        (if (memq name (map slot-definition-name (class-direct-slots class)))
 	   ;; This slot is direct; create a new shared variable
-	   (let ((shared-variable (make-unbound)))
-	     (list (lambda (o)   shared-variable)
-		   (lambda (o v) (set! shared-variable v))))
+	   (make-closure-variable class)
 	   ;; Slot is inherited. Find its definition in superclass
 	   (let loop ((l (cdr (class-precedence-list class))))
 	     (let ((r (assoc name (slot-ref (car l) 'getters-n-setters))))
@@ -1141,9 +1169,7 @@
 
     ((#:each-subclass) ;; slot shared by instances of direct subclass.
      ;; (Thomas Buerger, April 1998)
-     (let ((shared-variable (make-unbound)))
-       (list (lambda (o)   shared-variable)
-	     (lambda (o v) (set! shared-variable v)))))
+     (make-closure-variable class))
 
     ((#:virtual) ;; No allocation
      ;; slot-ref and slot-set! function must be given by the user
@@ -1156,9 +1182,17 @@
        (list get set)))
     (else    (next-method))))
 
+(define (make-closure-variable class)
+  (let ((shared-variable (make-unbound)))
+    (list (lambda (o) shared-variable)
+	  (lambda (o v) (set! shared-variable v)))))
+
 (define-method compute-get-n-set ((o <object>) s)
   (goops-error "Allocation \"%S\" is unknown" (slot-definition-allocation s)))
 
+(define-method compute-slots ((class <class>))
+  (%compute-slots class))
+
 ;;;
 ;;; {Initialize}
 ;;;
@@ -1180,7 +1214,7 @@
     (slot-set! class 'cpl		(compute-cpl class))
     (slot-set! class 'redefined		#f)
     (slot-set! class 'environment	env)
-    (let ((slots (%compute-slots class)))
+    (let ((slots (compute-slots class)))
       (slot-set! class 'slots	  	  slots)
       (slot-set! class 'nfields	  	  0)
       (slot-set! class 'getters-n-setters (compute-getters-n-setters class 
@@ -1202,10 +1236,7 @@
     (%inherit-magic! class supers)
 
     ;; Set the layout slot
-    (%prep-layout! class)
-
-    ;; Set the struct print closure
-    (slot-set! class 'print write-object)))
+    (%prep-layout! class)))
 
 (define object-procedure-tags
   '(utag_closure utag_subr_1 utag_subr_2 utag_subr3 utag_lsubr_2))
@@ -1290,12 +1321,19 @@
 			       slot
 			       (apply init '()))))))
 	      (map slot-definition-name (class-slots new-class)))
-    ;; Exchange old an new instance in place to keep pointers valids
+    ;; Exchange old and new instance in place to keep pointers valid
     (%modify-instance old-instance new-instance)
+    ;; Allow class specific updates of instances (which now are swapped)
+    (update-instance-for-different-class new-instance old-instance)
     old-instance))
 
 
-;; *fixme* Doesn't class-of cause an implicit call to change-object-class?
+(define-method update-instance-for-different-class ((old-instance <object>)
+						    (new-instance
+						     <object>))
+  ;;not really important what we do, we just need a default method
+  new-instance)
+
 (define-method change-class ((old-instance <object>) (new-class <class>))
   (change-object-class old-instance (class-of old-instance) new-class))
 
